# Audio dataset and corresponding functions taken from Patter
# https://github.com/ryanleary/patter
# TODO: review, and copyright and fix/add comments
import os

import kaldi_io
import torch
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.asr.parts import collections, parsers


def seq_collate_fn(batch, token_pad_value=0):
    """collate batch of audio sig, audio len, tokens, tokens len

    Args:
        batch (Optional[FloatTensor], Optional[LongTensor], LongTensor,
               LongTensor):  A tuple of tuples of signal, signal lengths,
               encoded tokens, and encoded tokens length.  This collate func
               assumes the signals are 1d torch tensors (i.e. mono audio).

    """
    _, audio_lengths, _, tokens_lengths = zip(*batch)
    max_audio_len = 0
    has_audio = audio_lengths[0] is not None
    if has_audio:
        max_audio_len = max(audio_lengths).item()
    max_tokens_len = max(tokens_lengths).item()

    audio_signal, tokens = [], []
    for sig, sig_len, tokens_i, tokens_i_len in batch:
        if has_audio:
            sig_len = sig_len.item()
            if sig_len < max_audio_len:
                pad = (0, max_audio_len - sig_len)
                sig = torch.nn.functional.pad(sig, pad)
            audio_signal.append(sig)
        tokens_i_len = tokens_i_len.item()
        if tokens_i_len < max_tokens_len:
            pad = (0, max_tokens_len - tokens_i_len)
            tokens_i = torch.nn.functional.pad(tokens_i, pad, value=token_pad_value)
        tokens.append(tokens_i)

    if has_audio:
        audio_signal = torch.stack(audio_signal)
        audio_lengths = torch.stack(audio_lengths)
    else:
        audio_signal, audio_lengths = None, None
    tokens = torch.stack(tokens)
    tokens_lengths = torch.stack(tokens_lengths)

    return audio_signal, audio_lengths, tokens, tokens_lengths


def fixed_seq_collate_fn(batch, fixed_length=16000):
    """collate batch of audio sig, audio len, tokens, tokens len

    Args:
        batch (Optional[FloatTensor], Optional[LongTensor], LongTensor,
               LongTensor):  A tuple of tuples of signal, signal lengths,
               encoded tokens, and encoded tokens length.  This collate func
               assumes the signals are 1d torch tensors (i.e. mono audio).
        fixed_length (Optional[int]): length of input signal to be considered

    """
    _, audio_lengths, _, tokens_lengths = zip(*batch)

    has_audio = audio_lengths[0] is not None
    fixed_length = min(fixed_length, max(audio_lengths))

    audio_signal, tokens = [], []
    for sig, sig_len, tokens_i, _ in batch:
        if has_audio:
            sig_len = sig_len.item()
            chunck_len = sig_len - fixed_length
            if chunck_len < 0:
                # pad = (0,fixed_length-sig_len)
                # signal = torch.nn.functional.pad(sig,pad)
                repeat = fixed_length // sig_len
                rem = fixed_length % sig_len
                sub = sig[-rem:] if rem > 0 else torch.tensor([])
                rep_sig = torch.cat(repeat * [sig])
                signal = torch.cat((rep_sig, sub))
                # print(sig_len,repeat,rem,len(sub),len(rep_sig),len(signal))
            else:
                start_idx = torch.randint(0, chunck_len, (1,)) if chunck_len else torch.tensor(0)
                end_idx = start_idx + fixed_length
                signal = sig[start_idx:end_idx]

            audio_signal.append(signal)
        tokens.append(tokens_i)

    if has_audio:
        audio_signal = torch.stack(audio_signal)
        audio_lengths = torch.stack(audio_lengths)
    else:
        audio_signal, audio_lengths = None, None
    tokens = torch.stack(tokens)
    tokens_lengths = torch.stack(tokens_lengths)

    return audio_signal, audio_lengths, tokens, tokens_lengths


def audio_seq_collate_fn(batch):
    """
    Collate a batch (iterable of (sample tensor, label tensor) tuples) into
    properly shaped data tensors
    :param batch:
    :return: inputs (batch_size, num_features, seq_length), targets,
    input_lengths, target_sizes
    """
    # sort batch by descending sequence length (for packed sequences later)
    batch.sort(key=lambda x: x[0].size(0), reverse=True)

    # init tensors we need to return
    inputs = []
    input_lengths = []
    target_sizes = []
    targets = []
    metadata = []

    # iterate over minibatch to fill in tensors appropriately
    for i, sample in enumerate(batch):
        input_lengths.append(sample[0].size(0))
        inputs.append(sample[0])
        target_sizes.append(len(sample[1]))
        targets.extend(sample[1])
        metadata.append(sample[2])
    targets = torch.tensor(targets, dtype=torch.long)
    inputs = torch.stack(inputs)
    input_lengths = torch.stack(input_lengths)
    target_sizes = torch.stack(target_sizes)

    return inputs, targets, input_lengths, target_sizes, metadata


class AudioDataset(Dataset):
    """
    Dataset that loads tensors via a json file containing paths to audio
    files, transcripts, and durations (in seconds). Each new line is a
    different sample. Example below:

    {"audio_filepath": "/path/to/audio.wav", "text_filepath":
    "/path/to/audio.txt", "duration": 23.147}
    ...
    {"audio_filepath": "/path/to/audio.wav", "text": "the
    transcription", "offset": 301.75, "duration": 0.82, "utt":
    "utterance_id", "ctm_utt": "en_4156", "side": "A"}

    Args:
        manifest_filepath: Path to manifest json as described above. Can
            be comma-separated paths.
        labels: String containing all the possible characters to map to
        featurizer: Initialized featurizer class that converts paths of
            audio to feature tensors
        max_duration: If audio exceeds this length, do not include in dataset
        min_duration: If audio is less than this length, do not include
            in dataset
        max_utts: Limit number of utterances
        blank_index: blank character index, default = -1
        unk_index: unk_character index, default = -1
        normalize: whether to normalize transcript text (default): True
        bos_id: Id of beginning of sequence symbol to append if not None
        eos_id: Id of end of sequence symbol to append if not None
        load_audio: Boolean flag indicate whether do or not load audio
        add_misc: True if add adiditional info dict.
    """

    def __init__(
        self,
        manifest_filepath,
        labels,
        featurizer,
        max_duration=None,
        min_duration=None,
        max_utts=0,
        blank_index=-1,
        unk_index=-1,
        normalize=True,
        trim=False,
        bos_id=None,
        eos_id=None,
        load_audio=True,
        parser='en',
        add_misc=False,
    ):
        self.collection = collections.ASRAudioText(
            manifests_files=manifest_filepath.split(','),
            parser=parsers.make_parser(
                labels=labels, name=parser, unk_id=unk_index, blank_id=blank_index, do_normalize=normalize,
            ),
            min_duration=min_duration,
            max_duration=max_duration,
            max_number=max_utts,
        )

        self.featurizer = featurizer
        self.trim = trim
        self.eos_id = eos_id
        self.bos_id = bos_id
        self.load_audio = load_audio
        self._add_misc = add_misc

    def __getitem__(self, index):
        sample = self.collection[index]
        if self.load_audio:
            offset = sample.offset

            if offset is None:
                offset = 0

            features = self.featurizer.process(
                sample.audio_file, offset=offset, duration=sample.duration, trim=self.trim,
            )
            f, fl = features, torch.tensor(features.shape[0]).long()
        else:
            f, fl = None, None

        t, tl = sample.text_tokens, len(sample.text_tokens)
        if self.bos_id is not None:
            t = [self.bos_id] + t
            tl += 1
        if self.eos_id is not None:
            t = t + [self.eos_id]
            tl += 1

        output = f, fl, torch.tensor(t).long(), torch.tensor(tl).long()

        if self._add_misc:
            misc = {}
            misc['id'] = sample.id
            misc['text_raw'] = sample.text_raw
            misc['speaker'] = sample.speaker
            output = (output, misc)

        return output

    def __len__(self):
        return len(self.collection)


class KaldiFeatureDataset(Dataset):
    """
    Dataset that provides basic Kaldi-compatible dataset loading. Assumes that
    the files `feats.scp`, `text`, and (optionally) `utt2dur` exist, as well
    as the .ark files that `feats.scp` points to.

    Args:
        kaldi_dir: Path to directory containing the aforementioned files.
        labels: All possible characters to map to.
        min_duration: If audio is shorter than this length, drop it. Only
            available if the `utt2dur` file exists.
        max_duration: If audio is longer than this length, drop it. Only
            available if the `utt2dur` file exists.
        max_utts: Limits the number of utterances.
        unk_index: unk_character index, default = -1
        blank_index: blank character index, default = -1
        normalize: whether to normalize transcript text. Defaults to True.
        eos_id: Id of end of sequence symbol to append if not None.
    """

    def __init__(
        self,
        kaldi_dir,
        labels,
        min_duration=None,
        max_duration=None,
        max_utts=0,
        unk_index=-1,
        blank_index=-1,
        normalize=True,
        eos_id=None,
    ):
        self.eos_id = eos_id
        self.unk_index = unk_index
        self.blank_index = blank_index
        self.labels_map = {label: i for i, label in enumerate(labels)}

        data = []
        duration = 0.0
        filtered_duration = 0.0

        # Read Kaldi features (MFCC, PLP) using feats.scp
        feats_path = os.path.join(kaldi_dir, 'feats.scp')
        id2feats = {utt_id: torch.from_numpy(feats) for utt_id, feats in kaldi_io.read_mat_scp(feats_path)}

        # Get durations, if utt2dur exists
        utt2dur_path = os.path.join(kaldi_dir, 'utt2dur')
        id2dur = {}
        if os.path.exists(utt2dur_path):
            with open(utt2dur_path, 'r') as f:
                for line in f:
                    utt_id, dur = line.split()
                    id2dur[utt_id] = float(dur)
        elif max_duration or min_duration:
            raise ValueError(
                f"KaldiFeatureDataset max_duration or min_duration is set but"
                f" utt2dur file not found in {kaldi_dir}."
            )
        else:
            logging.info(
                f"Did not find utt2dur when loading data from " f"{kaldi_dir}. Skipping dataset duration calculations."
            )

        # Match transcripts to features
        text_path = os.path.join(kaldi_dir, 'text')
        parser = parsers.make_parser(labels, 'en', unk_id=unk_index, blank_id=self.blank_index)
        with open(text_path, 'r') as f:
            for line in f:
                split_idx = line.find(' ')
                utt_id = line[:split_idx]

                audio_features = id2feats.get(utt_id)

                if audio_features is not None:

                    text = line[split_idx:].strip()
                    if normalize:
                        # TODO: WTF?
                        text = parser._normalize(text)

                    dur = id2dur[utt_id] if id2dur else None

                    # Filter by duration if specified & utt2dur exists
                    if min_duration and dur < min_duration:
                        filtered_duration += dur
                        continue
                    if max_duration and dur > max_duration:
                        filtered_duration += dur
                        continue

                    sample = {
                        'utt_id': utt_id,
                        'text': text,
                        'tokens': parser(text),
                        'audio': audio_features.t(),
                        'duration': dur,
                    }

                    data.append(sample)
                    duration += dur

                    if max_utts > 0 and len(data) >= max_utts:
                        logging.warning(f"Stop parsing due to max_utts ({max_utts})")
                        break

        if id2dur:
            # utt2dur durations are in seconds
            logging.info(
                f"Dataset loaded with {duration / 3600 : .2f} hours. "
                f"Filtered {filtered_duration / 3600 : .2f} hours."
            )

        self.data = data

    def __getitem__(self, index):
        sample = self.data[index]
        f = sample['audio']
        fl = torch.tensor(f.shape[1]).long()
        t, tl = sample['tokens'], len(sample['tokens'])

        if self.eos_id is not None:
            t.append(self.eos_id)
            tl += 1

        return f, fl, torch.tensor(t).long(), torch.tensor(tl).long()

    def __len__(self):
        return len(self.data)


class TranscriptDataset(Dataset):
    """A dataset class that reads and returns the text of a file.

    Args:
        path: (str) Path to file with newline separate strings of text
        labels (list): List of string labels to use when to str2int translation
        eos_id (int): Label position of end of string symbol
    """

    def __init__(self, path, labels, bos_id=None, eos_id=None, lowercase=True):
        parser = parsers.make_parser(labels, do_lowercase=lowercase)
        self.texts = collections.FromFileText(path, parser=parser)

        self.bos_id = bos_id
        self.eos_id = eos_id

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, item):
        tokenized_text = self.texts[item].tokens
        if self.bos_id:
            tokenized_text = [self.bos_id] + tokenized_text
        if self.eos_id:
            tokenized_text = tokenized_text + [self.eos_id]
        return (
            torch.tensor(tokenized_text, dtype=torch.long),
            torch.tensor(len(tokenized_text), dtype=torch.long),
        )


class AudioLabelDataset(Dataset):
    """
    Dataset that loads tensors via a json file containing paths to audio
    files, command class, and durations (in seconds). Each new line is a
    different sample. Example below:

    {"audio_filepath": "/path/to/audio.wav", "label":
    "label", "duration": 23.147}
    ...
    {"audio_filepath": "/path/to/audio.wav", "label": "label",
    "offset": 301.75, "duration": 0.82}

    Args:
        manifest_filepath: Path to manifest json as described above. Can
            be comma-separated paths.
        labels (Optional[list]): String containing all the possible labels to map to
            if None then automatically picks from ASRSpeechLabel collection.
        featurizer: Initialized featurizer class that converts paths of
            audio to feature tensors
        max_duration: If audio exceeds this length, do not include in dataset
        min_duration: If audio is less than this length, do not include
            in dataset
        trim: Boolean flag whether to trim the audio
        load_audio: Boolean flag indicate whether do or not load audio
    """

    def __init__(
        self,
        manifest_filepath,
        featurizer,
        labels=None,
        max_duration=None,
        min_duration=None,
        trim=False,
        load_audio=True,
    ):
        self.collection = collections.ASRSpeechLabel(
            manifests_files=manifest_filepath.split(','), min_duration=min_duration, max_duration=max_duration,
        )

        self.featurizer = featurizer
        self.trim = trim
        self.load_audio = load_audio

        self.labels = labels if labels else self.collection.uniq_labels
        self.num_commands = len(self.labels)

        self.label2id, self.id2label = {}, {}
        for label_id, label in enumerate(self.labels):
            self.label2id[label] = label_id
            self.id2label[label_id] = label

        for idx in range(len(self.labels[:5])):
            logging.debug(" label id {} and its mapped label {}".format(idx, self.id2label[idx]))

    def __getitem__(self, index):
        sample = self.collection[index]
        if self.load_audio:
            offset = sample.offset

            if offset is None:
                offset = 0

            features = self.featurizer.process(
                sample.audio_file, offset=offset, duration=sample.duration, trim=self.trim
            )
            f, fl = features, torch.tensor(features.shape[0]).long()
        else:
            f, fl = None, None

        t = self.label2id[sample.label]
        tl = 1  # For compatibility with collate_fn used later

        return f, fl, torch.tensor(t).long(), torch.tensor(tl).long()

    def __len__(self):
        return len(self.collection)



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.core.neural_types import LogprobsType, NeuralType, PredictionsType
from nemo.utils.decorators import add_port_docs


class GreedyCTCDecoder(NonTrainableNM):
    """
    Greedy decoder that computes the argmax over a softmax distribution
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns:
            Definitions of module input ports.
        """
        return {"log_probs": NeuralType(('B', 'T', 'D'), LogprobsType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns:
            Definitions of module output ports.
        """
        return {"predictions": NeuralType(('B', 'T'), PredictionsType())}

    def __init__(self):
        super().__init__()

    def forward(self, log_probs):
        argmx = log_probs.argmax(dim=-1, keepdim=False)
        return argmx



# Copyright (c) 2019 NVIDIA Corporation
from collections import UserDict


def pad_to(x, k=8):
    """Pad int value up to divisor of k.

    Examples:
        >>> pad_to(31, 8)
        32

    """

    return x + (x % k > 0) * (k - x % k)


class Config(UserDict):
    MAIN_SECTIONS = ('optimization', 'input', 'target', 'inference')

    def __init__(self, data, **kwargs):
        super().__init__(data, **kwargs)

        for section in self.MAIN_SECTIONS:
            if section not in self.data:
                self.data[section] = {}



# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import argparse

import nemo.utils.argparse as nm_argparse
from nemo.collections.cv.modules.data_layers import CIFAR10DataLayer
from nemo.collections.cv.modules.losses import NLLLoss
from nemo.collections.cv.modules.non_trainables import NonLinearity, ReshapeTensor
from nemo.collections.cv.modules.trainables import ConvNetEncoder, FeedForwardNetwork
from nemo.core import DeviceType, NeuralGraph, NeuralModuleFactory, OperationMode, SimpleLossLoggerCallback
from nemo.utils import logging

if __name__ == "__main__":
    # Create the default parser.
    parser = argparse.ArgumentParser(parents=[nm_argparse.NemoArgParser()], conflict_handler='resolve')
    # Parse the arguments
    args = parser.parse_args()

    # Instantiate Neural Factory.
    nf = NeuralModuleFactory(local_rank=args.local_rank, placement=DeviceType.CPU)

    # Data layer for training.
    cifar10_dl = CIFAR10DataLayer(train=True)
    # The "model".
    cnn = ConvNetEncoder(input_depth=3, input_height=32, input_width=32)
    reshaper = ReshapeTensor(input_sizes=[-1, 16, 2, 2], output_sizes=[-1, 64])
    ffn = FeedForwardNetwork(input_size=64, output_size=10, dropout_rate=0.1)
    nl = NonLinearity(type="logsoftmax", sizes=[-1, 10])
    # Loss.
    nll_loss = NLLLoss()

    # Create a training graph.
    with NeuralGraph(operation_mode=OperationMode.training) as training_graph:
        _, img, tgt = cifar10_dl()
        feat_map = cnn(inputs=img)
        res_img = reshaper(inputs=feat_map)
        logits = ffn(inputs=res_img)
        pred = nl(inputs=logits)
        loss = nll_loss(predictions=pred, targets=tgt)
        # Set output - that output will be used for training.
        training_graph.outputs["loss"] = loss

    # Display the graph summmary.
    logging.info(training_graph.summary())

    # SimpleLossLoggerCallback will print loss values to console.
    callback = SimpleLossLoggerCallback(
        tensors=[loss], print_func=lambda x: logging.info(f'Training Loss: {str(x[0].item())}')
    )

    # Invoke the "train" action.
    nf.train(
        training_graph=training_graph,
        callbacks=[callback],
        optimization_params={"num_epochs": 10, "lr": 0.001},
        optimizer="adam",
    )



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np

from nemo import logging

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']
GLOBAL_KEYS = ["eval_loss", "sys"]


def eval_iter_callback(tensors, global_vars):
    for key in GLOBAL_KEYS:
        if key not in global_vars.keys():
            global_vars[key] = []

    for kv, v in tensors.items():
        if "loss" in kv:
            for eval_loss in v:
                global_vars["eval_loss"].append(eval_loss.item())


def eval_epochs_done_callback(global_vars):
    eval_loss = np.mean(global_vars["eval_loss"])
    eval_ppl = np.exp(eval_loss)

    logging.info("------------------------------------------------------")
    logging.info("Eval loss: {0}".format(np.round(eval_loss, 3)))
    logging.info("Eval  ppl: {0}".format(np.round(eval_ppl, 3)))
    logging.info("------------------------------------------------------")
    for key in GLOBAL_KEYS:
        global_vars[key] = []
    return dict({"Eval_loss": eval_loss, "Eval_ppl": eval_ppl})



# Copyright (c) 2019 NVIDIA Corporation
# Requires Baidu's CTC decoders from
# https://github.com/PaddlePaddle/DeepSpeech/decoders/swig

import torch

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.core import DeviceType
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs
from nemo.utils.helpers import get_cuda_device


class BeamSearchDecoderWithLM(NonTrainableNM):
    """Neural Module that does CTC beam search with a n-gram language model.

    It takes a batch of log_probabilities. Note the bigger the batch, the
    better as proccessing is parallelized. Outputs a list of size batch_size.
    Each element in the list is a list of size beam_search, and each element
    in that list is a tuple of (final_log_prob, hyp_string).

    Args:
        vocab (list): List of characters that can be output by the ASR model. For Jasper, this is the 28 character set
            {a-z '}. The CTC blank symbol is automatically added later for models using ctc.
        beam_width (int): Size of beams to keep and expand upon. Larger beams result in more accurate but slower
            predictions
        alpha (float): The amount of importance to place on the n-gram language model. Larger alpha means more
            importance on the LM and less importance on the acoustic model (Jasper).
        beta (float): A penalty term given to longer word sequences. Larger beta will result in shorter sequences.
        lm_path (str): Path to n-gram language model
        num_cpus (int): Number of cpus to use
        cutoff_prob (float): Cutoff probability in vocabulary pruning, default 1.0, no pruning
        cutoff_top_n (int): Cutoff number in pruning, only top cutoff_top_n characters with highest probs in
            vocabulary will be used in beam search, default 40.
        input_tensor (bool): Set to True if you intend to pass pytorch Tensors, set to False if you intend to pass
            numpy arrays.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "log_probs": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(ChannelTag),}),
            # "log_probs_length": NeuralType({0: AxisType(BatchTag)}),
            "log_probs": NeuralType(('B', 'T', 'D'), LogprobsType()),
            "log_probs_length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        predictions:
            NeuralType(None)
        """
        # return {"predictions": NeuralType(VoidType())}
        return {"predictions": NeuralType(('B', 'T'), PredictionsType())}

    def __init__(
        self, vocab, beam_width, alpha, beta, lm_path, num_cpus, cutoff_prob=1.0, cutoff_top_n=40, input_tensor=True
    ):

        try:
            from ctc_decoders import Scorer
            from ctc_decoders import ctc_beam_search_decoder_batch
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "BeamSearchDecoderWithLM requires the "
                "installation of ctc_decoders "
                "from nemo/scripts/install_decoders.py"
            )

        super().__init__()
        # Override the default placement from neural factory and set placement/device to be CPU.
        self._placement = DeviceType.CPU
        self._device = get_cuda_device(self._placement)

        if self._factory.world_size > 1:
            raise ValueError("BeamSearchDecoderWithLM does not run in distributed mode")

        self.scorer = Scorer(alpha, beta, model_path=lm_path, vocabulary=vocab)
        self.beam_search_func = ctc_beam_search_decoder_batch
        self.vocab = vocab
        self.beam_width = beam_width
        self.num_cpus = num_cpus
        self.cutoff_prob = cutoff_prob
        self.cutoff_top_n = cutoff_top_n
        self.input_tensor = input_tensor

    def forward(self, log_probs, log_probs_length):
        probs_list = log_probs
        if self.input_tensor:
            probs = torch.exp(log_probs)
            probs_list = []
            for i, prob in enumerate(probs):
                probs_list.append(prob[: log_probs_length[i], :])
        res = self.beam_search_func(
            probs_list,
            self.vocab,
            beam_size=self.beam_width,
            num_processes=self.num_cpus,
            ext_scoring_func=self.scorer,
            cutoff_prob=self.cutoff_prob,
            cutoff_top_n=self.cutoff_top_n,
        )
        return [res]



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import itertools

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils import (
    fill_class_weights,
    get_freq_weights,
    get_label_stats,
    if_exist,
)

__all__ = ['JointIntentSlotDataDesc']


class JointIntentSlotDataDesc:
    """ Convert the raw data to the standard format supported by
    JointIntentSlotDataset.

    By default, the None label for slots is 'O'.

    JointIntentSlotDataset requires two files:

        input_file: file to sequence + label.
            the first line is header (sentence [tab] label)
            each line should be [sentence][tab][label]

        slot_file: file to slot labels, each line corresponding to
            slot labels for a sentence in input_file. No header.

    To keep the mapping from label index to label consistent during
    training and inferencing, we require the following files:
        dicts.intents.csv: each line is an intent. The first line
            corresponding to the 0 intent label, the second line
            corresponding to the 1 intent label, and so on.

        dicts.slots.csv: each line is a slot. The first line
            corresponding to the 0 slot label, the second line
            corresponding to the 1 slot label, and so on.

    Args:
        data_dir (str): the directory of the dataset
        none_slot_label (str): the label for slots that aren't identified
            defaulted to 'O'
        pad_label (int): the int used for padding. If set to -1,
             it'll be set to the whatever the None label is.
    """

    def __init__(self, data_dir, none_slot_label='O', pad_label=-1):
        if not if_exist(data_dir, ['dict.intents.csv', 'dict.slots.csv']):
            raise FileNotFoundError(
                "Make sure that your data follows the standard format "
                "supported by JointIntentSlotDataset. Your data must "
                "contain dict.intents.csv and dict.slots.csv."
            )

        self.data_dir = data_dir
        self.intent_dict_file = self.data_dir + '/dict.intents.csv'
        self.slot_dict_file = self.data_dir + '/dict.slots.csv'

        self.intents_label_ids = JointIntentSlotDataDesc.label2idx(self.intent_dict_file)
        self.num_intents = len(self.intents_label_ids)
        self.slots_label_ids = JointIntentSlotDataDesc.label2idx(self.slot_dict_file)
        self.num_slots = len(self.slots_label_ids)

        infold = self.data_dir
        for mode in ['train', 'test', 'dev']:
            if not if_exist(self.data_dir, [f'{mode}.tsv']):
                logging.info(f' Stats calculation for {mode} mode' f' is skipped as {mode}.tsv was not found.')
                continue
            logging.info(f' Stats calculating for {mode} mode...')
            slot_file = f'{self.data_dir}/{mode}_slots.tsv'
            with open(slot_file, 'r') as f:
                slot_lines = f.readlines()

            input_file = f'{self.data_dir}/{mode}.tsv'
            with open(input_file, 'r') as f:
                input_lines = f.readlines()[1:]  # Skipping headers at index 0

            if len(slot_lines) != len(input_lines):
                raise ValueError(
                    "Make sure that the number of slot lines match the "
                    "number of intent lines. There should be a 1-1 "
                    "correspondence between every slot and intent lines."
                )

            dataset = list(zip(slot_lines, input_lines))

            raw_slots, raw_intents = [], []
            for slot_line, input_line in dataset:
                slot_list = [int(slot) for slot in slot_line.strip().split()]
                raw_slots.append(slot_list)
                parts = input_line.strip().split()
                raw_intents.append(int(parts[-1]))

            logging.info(f'Three most popular intents in {mode} mode:')
            total_intents, intent_label_freq, max_id = get_label_stats(
                raw_intents, infold + f'/{mode}_intent_stats.tsv'
            )

            merged_slots = itertools.chain.from_iterable(raw_slots)
            logging.info(f'Three most popular slots in {mode} mode:')
            slots_total, slots_label_freq, max_id = get_label_stats(merged_slots, infold + f'/{mode}_slot_stats.tsv')

            logging.info(f'Total Number of Intents: {total_intents}')
            logging.info(f'Intent Label Frequencies: {intent_label_freq}')
            logging.info(f'Total Number of Slots: {slots_total}')
            logging.info(f'Slots Label Frequencies: {slots_label_freq}')

            if mode == 'train':
                intent_weights_dict = get_freq_weights(intent_label_freq)
                logging.info(f'Intent Weights: {intent_weights_dict}')
                slot_weights_dict = get_freq_weights(slots_label_freq)
                logging.info(f'Slot Weights: {slot_weights_dict}')

        self.intent_weights = fill_class_weights(intent_weights_dict, self.num_intents - 1)
        self.slot_weights = fill_class_weights(slot_weights_dict, self.num_slots - 1)

        if pad_label != -1:
            self.pad_label = pad_label
        else:
            if none_slot_label not in self.slots_label_ids:
                raise ValueError(f'none_slot_label {none_slot_label} not ' f'found in {self.slot_dict_file}.')
            self.pad_label = self.slots_label_ids[none_slot_label]

    @staticmethod
    def label2idx(file):
        lines = open(file, 'r').readlines()
        lines = [line.strip() for line in lines if line.strip()]
        labels = {lines[i]: i for i in range(len(lines))}
        return labels



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['AxisKindAbstract', 'AxisKind', 'AxisType']

from enum import Enum
from typing import Optional


class AxisKindAbstract(Enum):
    """This is an abstract Enum to represents what does varying axis dimension mean.
    In practice, you will almost always use AxisKind Enum. This Enum should be inherited by
    your OWN Enum if you aren't satisfied with AxisKind. Then your own Enum can be used
    instead of AxisKind."""

    pass


class AxisKind(AxisKindAbstract):
    """This Enum represents what does varying axis dimension mean.
    For example, does this dimension correspond to width, batch, time, etc.
    The "Dimension" and "Channel" kinds are the same and used to represent
    a general axis. "Any" axis will accept any axis kind fed to it.
    """

    Batch = 0
    Time = 1
    Dimension = 2
    Channel = 2
    Width = 3
    Height = 4
    Any = 5
    Sequence = 6

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(self.name).lower()

    @staticmethod
    def from_str(label):
        """Returns AxisKind instance based on short string representation"""
        _label = label.lower().strip()
        if _label == "b" or _label == "n" or _label == "batch":
            return AxisKind.Batch
        elif _label == "t" or _label == "time":
            return AxisKind.Time
        elif _label == "d" or _label == "c" or _label == "channel":
            return AxisKind.Dimension
        elif _label == "w" or _label == "width":
            return AxisKind.Width
        elif _label == "h" or _label == "height":
            return AxisKind.Height
        elif _label == "any":
            return AxisKind.Any
        else:
            raise ValueError(f"Can't create AxisKind from {label}")


class AxisType(object):
    """This class represents axis semantics and (optionally) it's dimensionality
       Args:
           kind (AxisKindAbstract): what kind of axis it is? For example Batch, Height, etc.
           size (int, optional): specify if the axis should have a fixed size. By default it is set to None and you
           typically do not want to set it for Batch and Time
           is_list (bool, default=False): whether this is a list or a tensor axis
    """

    def __init__(self, kind: AxisKindAbstract, size: Optional[int] = None, is_list=False):
        if size is not None and is_list:
            raise ValueError("The axis can't be list and have a fixed size")
        self.kind = kind
        self.size = size
        self.is_list = is_list

    def __repr__(self):
        if self.size is None:
            representation = str(self.kind)
        else:
            representation = f"{str(self.kind)}:{self.size}"
        if self.is_list:
            representation += "_listdim"
        return representation



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/train_and_predict.py
'''

from torch import nn

from nemo.backends.pytorch.nm import TrainableNM
from nemo.collections.nlp.utils.transformer_utils import transformer_weights_init
from nemo.core import ChannelType, EmbeddedTextType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['SGDEncoderNM']

ACTIVATIONS_F = {
    "tanh": nn.Tanh,
    "relu": nn.ReLU,
}


class SGDEncoderNM(TrainableNM):
    """
    Neural module which extracts the first token from the BERT representation of the utterance
    followed by a fully connected layer.

    Args:
        hidden_size (int): hidden size of the BERT model
        activation (str): activation function applied
        dropout (float): dropout ratio
    """

    @property
    @add_port_docs
    def input_ports(self):
        """
        Returns definitions of module input ports.
        hidden_states (float): BERT representation of the utterance
        """
        return {"hidden_states": NeuralType(('B', 'T', 'C'), ChannelType())}

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports.
        logits (float): First token of the BERT representation of the utterance followed by fc and dropout
        hidden_states (float) : BERT representation of the utterance with applied dropout
        """
        return {
            "logits": NeuralType(('B', 'T'), EmbeddedTextType()),
            "hidden_states": NeuralType(('B', 'T', 'C'), ChannelType()),
        }

    def __init__(self, hidden_size, activation='tanh', dropout=0.0, use_transformer_pretrained=True):
        super().__init__()
        self.fc = nn.Linear(hidden_size, hidden_size).to(self._device)

        if activation not in ACTIVATIONS_F:
            raise ValueError(f'{activation} is not in supported ' + '{ACTIVATIONS_F.keys()}')

        self.activation = ACTIVATIONS_F[activation]()
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        if use_transformer_pretrained:
            self.apply(lambda module: transformer_weights_init(module, xavier=False))
        # self.to(self._device) # sometimes this is necessary

    def forward(self, hidden_states):
        first_token_hidden_states = hidden_states[:, 0]
        logits = self.fc(first_token_hidden_states)
        logits = self.activation(logits)
        logits = self.dropout1(logits)
        return logits, self.dropout2(hidden_states)



# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for GLUE tasks
Some transformer of this code were adapted from the HuggingFace library at
https://github.com/huggingface/transformers
"""
import os
import random

import numpy as np
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import f1_score, matthews_corrcoef

from nemo import logging
from nemo.collections.nlp.utils.callback_utils import list2str, tensor2list

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars):
    if "all_preds" not in global_vars.keys():
        global_vars["all_preds"] = []
    if "all_labels" not in global_vars.keys():
        global_vars["all_labels"] = []

    logits_lists = []
    preds_lists = []
    labels_lists = []

    for kv, v in tensors.items():
        # for GLUE classification tasks
        if 'logits' in kv:
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits_lists.append(tensor2list(logit_tensor))
        # for GLUE STS-B task (regression)
        elif 'preds' in kv:
            for v_tensor in v:
                for pred_tensor in v_tensor:
                    preds_lists.append(tensor2list(pred_tensor))
        if 'labels' in kv:
            for v_tensor in v:
                for label_tensor in v_tensor:
                    labels_lists.append(tensor2list(label_tensor))

    if len(logits_lists) > 0:
        preds = list(np.argmax(np.asarray(logits_lists), 1))
    elif len(preds_lists) > 0:
        preds = list(np.squeeze(np.asarray(preds_lists)))

    global_vars["all_preds"].extend(preds)
    global_vars["all_labels"].extend(labels_lists)


def eval_epochs_done_callback(global_vars, output_dir, task_name):
    labels = np.asarray(global_vars['all_labels'])
    preds = np.asarray(global_vars['all_preds'])

    # print predictions and labels for a small random subset of data
    sample_size = 20
    i = 0
    if preds.shape[0] > sample_size + 1:
        i = random.randint(0, preds.shape[0] - sample_size - 1)

    logging.info("Task name: %s" % task_name.upper())
    logging.info("Sampled preds: [%s]" % list2str(preds[i : i + sample_size]))
    logging.info("Sampled labels: [%s]" % list2str(labels[i : i + sample_size]))

    results = compute_metrics(task_name, preds, labels)

    os.makedirs(output_dir, exist_ok=True)
    with open(os.path.join(output_dir, task_name + '.txt'), 'w') as f:
        f.write('labels\t' + list2str(labels) + '\n')
        f.write('preds\t' + list2str(preds) + '\n')

    logging.info(results)

    return results


def accuracy(preds, labels):
    return {"acc": (preds == labels).mean()}


def acc_and_f1(preds, labels):
    accuracy = (preds == labels).mean()
    f1 = f1_score(y_true=labels, y_pred=preds)
    return {"acc": accuracy, "f1": f1, "acc_and_f1": (accuracy + f1) / 2}


def mcc(preds, labels):
    return {"mcc": matthews_corrcoef(labels, preds)}


def pearson_and_spearman(preds, labels):
    pearson_corr = pearsonr(preds, labels)[0]
    spearman_corr = spearmanr(preds, labels)[0]
    return {"pearson": pearson_corr, "spearmanr": spearman_corr, "corr": (pearson_corr + spearman_corr) / 2}


def compute_metrics(task_name, preds, labels):
    if len(preds) != len(labels):
        raise ValueError("Predictions and labels must have the same lenght")

    metric_fn = accuracy
    if task_name == 'cola':
        metric_fn = mcc
    elif task_name in ['mrpc', 'qqp']:
        metric_fn = acc_and_f1
    elif task_name == 'sts-b':
        metric_fn = pearson_and_spearman

    return metric_fn(preds, labels)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.backends.pytorch import DataLayerNM
from nemo.collections.nlp.data.datasets import *

__all__ = ['TextDataLayer']


class TextDataLayer(DataLayerNM):
    """
    Generic Text Data Layer NM which wraps PyTorch's dataset

    Args:
        dataset_type (Dataset): type of dataset used for this datalayer
        dataset_params (dict): all the params for the dataset
        batch_size (int): sequence batch size
        shuffle (bool): whether to shuffle data
    """

    def __init__(self, dataset_type, dataset_params, batch_size, shuffle=False, num_workers=-1, pin_memory=False):
        super().__init__()
        self._dataset = dataset_type(**dataset_params)
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._pin_memory = pin_memory
        if num_workers >= 0:
            self._num_workers = num_workers

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None



# Copyright (c) 2019 NVIDIA Corporation
import torch


class LinearNorm(torch.nn.Module):
    def __init__(self, in_dim, out_dim, bias=True, w_init_gain='linear'):
        super(LinearNorm, self).__init__()
        self.linear_layer = torch.nn.Linear(in_dim, out_dim, bias=bias)

        torch.nn.init.xavier_uniform_(self.linear_layer.weight, gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, x):
        return self.linear_layer(x)


class ConvNorm(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=1,
        stride=1,
        padding=None,
        dilation=1,
        bias=True,
        w_init_gain='linear',
    ):
        super(ConvNorm, self).__init__()
        if padding is None:
            assert kernel_size % 2 == 1
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        torch.nn.init.xavier_uniform_(self.conv.weight, gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, signal):
        conv_signal = self.conv(signal)
        return conv_signal


def get_mask_from_lengths(lengths, max_len=None):
    if not max_len:
        max_len = torch.max(lengths).item()
    ids = torch.arange(0, max_len, out=torch.cuda.LongTensor(max_len))
    mask = (ids < lengths.unsqueeze(1)).bool()
    return mask



# Copyright (C) NVIDIA CORPORATION. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.****

import sys

from nemo.constants import NEMO_ENV_VARNAME_ENABLE_COLORING
from nemo.utils.env_var_parsing import get_envbool

__all__ = ["check_color_support", "to_unicode"]


def check_color_support():
    # Colors can be forced with an env variable
    if not sys.platform.lower().startswith("win") and get_envbool(NEMO_ENV_VARNAME_ENABLE_COLORING, False):
        return True


def to_unicode(value):
    """
    Converts a string argument to a unicode string.
    If the argument is already a unicode string or None, it is returned
    unchanged.  Otherwise it must be a byte string and is decoded as utf8.
    """
    try:
        if isinstance(value, (str, type(None))):
            return value

        if not isinstance(value, bytes):
            raise TypeError("Expected bytes, unicode, or None; got %r" % type(value))

        return value.decode("utf-8")

    except UnicodeDecodeError:
        return repr(value)



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'NeuralType',
    'NmTensor',
    'NeuralTypeError',
    'NeuralPortNameMismatchError',
    'NeuralPortNmTensorMismatchError',
]
import uuid
from typing import List, Optional, Tuple

from nemo.core.neural_types.axes import AxisKind, AxisType
from nemo.core.neural_types.comparison import NeuralTypeComparisonResult
from nemo.core.neural_types.elements import *
from nemo.utils.app_state import AppState
from nemo.utils.neural_graph.connection import Connection, StepModulePort


class NeuralType(object):
    """This is the main class which would represent neural type concept.
    nmTensors derives from this. It is used to represent *the types* of inputs and outputs.

    Args:
        axes (Optional[Tuple]): a tuple of AxisTypes objects representing the semantics of what varying each axis means
            You can use a short, string-based form here. For example: ('B', 'C', 'H', 'W') would correspond to an NCHW
            format frequently used in computer vision. ('B', 'T', 'D') is frequently used for signal processing and
            means [batch, time, dimension/channel].
        elements_type (ElementType): an instance of ElementType class representing the semantics of what is stored
            inside the tensor. For example: logits (LogitsType), log probabilities (LogprobType), etc.
        optional (bool): By default, this is false. If set to True, it would means that input to the port of this
            type can be optional.
    """

    def __str__(self):

        if self.axes is not None:
            return f"axes: {self.axes}; elements_type: {self.elements_type.__class__.__name__}"
        else:
            return f"axes: None; elements_type: {self.elements_type.__class__.__name__}"

    def __init__(self, axes: Optional[Tuple] = None, elements_type: ElementType = VoidType(), optional=False):
        if not isinstance(elements_type, ElementType):
            raise ValueError(
                f"elements_type of NeuralType must be an instance of a class derived from ElementType."
                f"Did you pass a class instead?"
            )
        self.elements_type = elements_type
        if axes is not None:
            NeuralType.__check_sanity(axes)
            axes_list = []
            for axis in axes:
                if isinstance(axis, str):
                    axes_list.append(AxisType(AxisKind.from_str(axis), None))
                elif isinstance(axis, AxisType):
                    axes_list.append(axis)
                else:
                    raise ValueError(f"axis type must be either str or AxisType instance")
            self.axes = tuple(axes_list)
        else:
            self.axes = None
        self.optional = optional

    def compare(self, second) -> NeuralTypeComparisonResult:
        """Performs neural type comparison of self with second. When you chain two modules' inputs/outputs via
        __call__ method, this comparison will be called to ensure neural type compatibility."""
        # First, handle dimensionality
        axes_a = self.axes
        axes_b = second.axes

        # "Big void" type
        if isinstance(self.elements_type, VoidType) and self.axes is None:
            return NeuralTypeComparisonResult.SAME

        if self.axes is None:
            if second.axes is None:
                return self.elements_type.compare(second.elements_type)
            else:
                return NeuralTypeComparisonResult.INCOMPATIBLE

        dimensions_pass = NeuralType.__compare_axes(axes_a, axes_b)
        element_comparison_result = self.elements_type.compare(second.elements_type)

        # SAME DIMS
        if dimensions_pass == 0:
            return element_comparison_result
        # TRANSPOSE_SAME DIMS
        elif dimensions_pass == 1:
            if element_comparison_result == NeuralTypeComparisonResult.SAME:
                return NeuralTypeComparisonResult.TRANSPOSE_SAME
            else:
                return NeuralTypeComparisonResult.INCOMPATIBLE
        # DIM_INCOMPATIBLE DIMS
        elif dimensions_pass == 2:
            if element_comparison_result == NeuralTypeComparisonResult.SAME:
                return NeuralTypeComparisonResult.DIM_INCOMPATIBLE
            else:
                return NeuralTypeComparisonResult.INCOMPATIBLE
        else:
            return NeuralTypeComparisonResult.INCOMPATIBLE

    def compare_and_raise_error(self, parent_type_name, port_name, second_object):
        """ Method compares definition of one type with another and raises an error if not compatible. """
        type_comatibility = self.compare(second_object)
        if (
            type_comatibility != NeuralTypeComparisonResult.SAME
            and type_comatibility != NeuralTypeComparisonResult.GREATER
        ):
            raise NeuralPortNmTensorMismatchError(
                parent_type_name, port_name, str(self), str(second_object.ntype), type_comatibility
            )

    @staticmethod
    def __check_sanity(axes):
        # check that list come before any tensor dimension
        are_strings = True
        for axis in axes:
            if not isinstance(axis, str):
                are_strings = False
            if isinstance(axis, str) and not are_strings:
                raise ValueError("Either use full class names or all strings")
        if are_strings:
            return
        checks_passed = True
        saw_tensor_dim = False
        for axis in axes:
            if not axis.is_list:
                saw_tensor_dim = True
            else:  # current axis is a list
                if saw_tensor_dim:  # which is preceded by tensor dim
                    checks_passed = False
        if not checks_passed:
            raise ValueError(
                "You have list dimension after Tensor dimension. All list dimensions must preceed Tensor dimensions"
            )

    @staticmethod
    def __compare_axes(axes_a, axes_b) -> int:
        """
        Compares axes_a and axes_b
        
        Args:
            axes_a: first axes tuple
            axes_b: second axes tuple

        Returns:
            0 - if they are exactly the same
            1 - if they are "TRANSPOSE_SAME"
            2 - if the are "DIM_INCOMPATIBLE"
            3 - if they are different
        """
        if axes_a is None and axes_b is None:
            return 0
        elif axes_a is None and axes_b is not None:
            return 3
        elif axes_a is not None and axes_b is None:
            return 3
        elif len(axes_a) != len(axes_b):
            return 3
        # After these ifs we know that len(axes_a) == len(axes_b)

        same = True
        kinds_a = {}
        kinds_b = {}
        for axis_a, axis_b in zip(axes_a, axes_b):
            kinds_a[axis_a.kind] = axis_a.size
            kinds_b[axis_b.kind] = axis_b.size
            if axis_a.kind == AxisKind.Any:
                same = True
            elif (
                axis_a.kind != axis_b.kind
                or axis_a.is_list != axis_b.is_list
                or (axis_a.size != axis_b.size and axis_a.size is not None)
            ):
                same = False
        if same:
            return 0
        else:
            # can be TRANSPOSE_SAME, DIM_INCOMPATIBLE
            if kinds_a.keys() == kinds_b.keys():
                for key, value in kinds_a.items():
                    if kinds_b[key] != value:
                        return 2
                return 1
            else:
                return 3


class NmTensor(NeuralType):
    """Class representing data which flows between NeuralModules' ports. It also has a type of NeuralType represented
    by inheriting from NeuralType object.
    """

    def __init__(self, producer, producer_args, output_port_name, ntype=None):
        """NmTensor constructor.

        Args:
          producer (NeuralModule): object which produced this
          producer_args (dict): a dictionary of port_name->NmTensor value
            of arguments which were sent to producer to create this
        """
        super(NmTensor, self).__init__(axes=ntype.axes, elements_type=ntype.elements_type, optional=ntype.optional)
        # producer is None: a special case present in some of the unit tests.
        if producer is None:
            self._producer_name = "None"
        else:
            self._producer_name = producer.name
        self._producer_args = producer_args
        self._output_port_name = output_port_name
        self._name = output_port_name
        self._output_port_name = output_port_name
        self._uuid = str(uuid.uuid4())
        # Remember step at which this tensor was created.
        self._step_number = AppState().active_graph.step_number
        # List of tuples (step number, module name, input port name)
        self._consumers = []
        AppState().tensor_names.register(self)

    @property
    def producer(self):
        """
        Returns:
            NeuralModule object which produced this NmTensor.
        """
        return AppState().modules[self._producer_name]

    @property
    def producer_name(self) -> str:
        """
        Returns:
            Name of the producer of the tensor.
        """
        return self._producer_name

    @property
    def producer_step_number(self) -> int:
        """
        Returns:
            Step number indicating when the tensor was produced. (It also indicates who produced the tensor.)
        """
        return self._step_number

    @property
    def producer_step_module_port(self) -> StepModulePort:
        """
        Returns:
            A tuple containing step number, module name and corresponding output port name.
        """
        return StepModulePort(self._step_number, self._producer_name, self._output_port_name)

    @property
    def consumers(self) -> List[StepModulePort]:
        """
        Returns:
            A list of tuples containing consumer step number, module name and corresponding input port names.
        """
        return self._consumers

    def add_consumer(self, step_module_port: StepModulePort):
        """
        Adds the "consumer" to tensor.

        Args:
            step_port: Step number, module name and module's input port.
        """
        self._consumers.append(step_module_port)

    @property
    def ntype(self):
        """
        Returns:
            Neural Type associated with this NmTensor.
        """
        return NeuralType(axes=self.axes, elements_type=self.elements_type, optional=self.optional)

    def connections(self):
        """
            "Serializes" the tensor to a list of connections (step/producer/port, step/consumer/port).
        """
        connections = []
        for con_mod_port in self._consumers:
            connections.append(Connection(self.producer_step_module_port, con_mod_port, self.ntype))
        return connections

    @property
    def producer_args(self):
        """
        Returns:
            A dictionary of port_name->NmTensor value of arguments which were sent to producer to create this object
        """
        return self._producer_args

    @property
    def name(self):
        """
        Returns:
            A NmTensor's name. By default, it is equal to self.output_port_name, but can be renamed using self.rename.
            Used in self.__str__() to obtain a string representation of a nmtensor.
        """
        return self._name

    @property
    def output_port_name(self):
        """
        Returns:
            The name of the output port it is associated with. This is used in DAG creation.
        """
        return self._output_port_name

    @property
    def unique_name(self):
        """Unique NMTensor name.
        It is composed of non-unique name (self.name) and uuid of NeuralModule which created this tensor.

        Returns:
            str: unique name
        """
        if self._producer_name is None:
            raise ValueError("This NmTensor does not have a unique name")
        return f"{self._output_port_name}~~~{self._producer_name}~~~{self._uuid}"

    def rename(self, new_name: str):
        """Renames the tensor from its old name to a new user-defined name for easy access within callbacks. Note,
        a tensor's unique_name is never changed. This simply adds a reference from new_name -> tensor.unique_name

        args:
            new_name (str): the new tensor's name.
        """
        AppState().tensor_names.rename_NmTensor(self, new_name)
        self._name = new_name

    def __str__(self):
        return self.name


class NeuralTypeError(Exception):
    """Base class for neural type related exceptions."""

    pass


class NeuralPortNameMismatchError(NeuralTypeError):
    """Exception raised when neural module is called with incorrect port
    names."""

    def __init__(self, input_port_name):
        self.message = "Wrong input port name: {0}".format(input_port_name)


class NeuralPortNmTensorMismatchError(NeuralTypeError):
    """Exception raised when a port is fed with a NmTensor of incompatible
    type."""

    def __init__(self, class_name, port_name, first_type, second_type, type_comatibility):
        self.message = "\nIn {}. \nPort: {} and a NmTensor it was fed are \n".format(class_name, port_name)
        self.message += "of incompatible neural types:\n\n{} \n\n and \n\n{}".format(first_type, second_type)
        self.message += "\n\nType comparison result: {}".format(type_comatibility)



# Copyright (c) 2019 NVIDIA Corporation
from typing import List, Optional

import editdistance
import sklearn
import torch


def __levenshtein(a: List[str], b: List[str]) -> int:
    """Calculates the Levenshtein distance between a and b."""
    return editdistance.eval(a, b)


def word_error_rate(hypotheses: List[str], references: List[str], use_cer=False) -> float:
    """
    Computes Average Word Error rate between two texts represented as
    corresponding lists of string. Hypotheses and references must have same
    length.
    Args:
      hypotheses: list of hypotheses
      references: list of references
      use_cer: bool, set True to enable cer
    Returns:
      (float) average word error rate
    """
    scores = 0
    words = 0
    if len(hypotheses) != len(references):
        raise ValueError(
            "In word error rate calculation, hypotheses and reference"
            " lists must have the same number of elements. But I got:"
            "{0} and {1} correspondingly".format(len(hypotheses), len(references))
        )
    for h, r in zip(hypotheses, references):
        if use_cer:
            h_list = list(h)
            r_list = list(r)
        else:
            h_list = h.split()
            r_list = r.split()
        words += len(r_list)
        scores += __levenshtein(h_list, r_list)
    if words != 0:
        wer = 1.0 * scores / words
    else:
        wer = float('inf')
    return wer


def classification_accuracy(
    logits: torch.Tensor, targets: torch.Tensor, top_k: Optional[List[int]] = None
) -> List[float]:
    """
    Computes the top-k classification accuracy provided with
    un-normalized logits of a model and ground truth targets.
    If top_k is not provided, defaults to top_1 accuracy.
    If top_k is provided as a list, then the values are sorted
    in ascending order.
    Args:
        logits: Un-normalized logits of a model. Softmax will be
            applied to these logits prior to computation of accuracy.
        targets: Vector of integers which represent indices of class
            labels.
        top_k: Optional list of integers in the range [1, max_classes].
    Returns:
        A list of length `top_k`, where each value represents top_i
        accuracy (i in `top_k`).
    """
    if top_k is None:
        top_k = [1]
    max_k = max(top_k)

    with torch.no_grad():
        _, predictions = logits.topk(max_k, dim=1, largest=True, sorted=True)
        predictions = predictions.t()
        correct = predictions.eq(targets.view(1, -1)).expand_as(predictions)

        results = []
        for k in top_k:
            correct_k = correct[:k].view(-1).float().mean().to('cpu').numpy()
            results.append(correct_k)

    return results


def classification_confusion_matrix(logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    """
    Computes the confusion matrix provided with 
    un-normalized logits of a model and ground truth targets.

    Args:
        logits: Un-normalized logits of a model. Softmax will be
            applied to these logits prior to computation of accuracy.
        targets: Vector of integers which represent indices of class
            labels.
    Returns:

        A tensor (n_classes, n_classes) of confusion matrix.
    """

    with torch.no_grad():
        _, predictions = logits.topk(1, dim=1, largest=True, sorted=True)
        predictions = predictions.t().squeeze()
        targets = targets.t().squeeze()

    return sklearn.metrics.confusion_matrix(targets, predictions, labels=range(logits.shape[1]))



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# Sets up everything needed to perform inference with TensorRT.
import os
import pickle
import sys
import time
import zlib
from collections import OrderedDict

import numpy as np

# Only initialize GPU after this runner is activated.
import pycuda.autoinit

# This import causes pycuda to automatically manage CUDA context creation and cleanup.
import pycuda.driver as cuda
import tensorrt as trt

from nemo import logging, logging_mode

logging.info("Using TensorRT {:}".format(trt.__version__))
logging.debug("Note: Using tensorrt from {:}".format(trt.__path__))

TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


def set_trt_logging_level(sev):
    global TRT_LOGGER
    if sev == logging.DEBUG:
        logging.min_severity = trt.Logger.INFO
    elif sev == logging.WARNING:
        logging.min_severity = trt.Logger.WARNING
    elif sev == logging.ERROR:
        logging.min_severity = trt.Logger.ERROR
    elif sev == logging.CRITICAL:
        logging.min_severity = trt.Logger.INTERNAL_ERROR


TRT_DYNAMIC_DIM = -1
DEFAULT_SHAPE_VALUE = 1


# Attempt to partially match output names. Returns None on failure
# Checks for exact matches and substring matches, falling back to index based matching.
def find_in_dict(name, map, index=None):
    if name in map:
        return name
    for key in map.keys():
        if name.lower() in key.lower() or key.lower() in name.lower():
            return key
    if index is not None and index >= 0 and index < len(map.keys()):
        return list(map.keys())[index]
    return None


def default_value(value, default):
    return value if value is not None else default


def is_dimension_dynamic(dim):
    return dim is None or dim <= 0


def is_shape_dynamic(shape):
    return any([is_dimension_dynamic(dim) for dim in shape])


def is_valid_shape_override(new_shape, original_shape):
    ranks_same = len(original_shape) == len(new_shape)
    overrides_valid = all(
        [odim == ndim or is_dimension_dynamic(odim) for odim, ndim in zip(original_shape, new_shape)]
    )
    return ranks_same and overrides_valid


def volume(obj):
    vol = 1
    for elem in obj:
        vol *= elem
    return vol


def compress(info):
    return zlib.compress(pickle.dumps(info))


def decompress(bytes):
    return pickle.loads(zlib.decompress(bytes))


def is_compressed(obj):
    return isinstance(obj, bytes)


def is_pickleable(obj):
    try:
        pickle.dumps(obj)
        return True
    except TypeError:
        return False


def pickle_load(path):
    with open(path, "rb") as f:
        return pickle.loads(f.read())


def pickle_save(path, obj):
    with open(path, "wb") as f:
        return f.write(pickle.dumps(obj))


# The maximum number of bytes that can be sent at once over a queue.
PIPE_MAX_SEND_BYTES = 1 << 31

# Attempts to send an object over the queue, compresses if needed. In the event the object cannot be sent, sends None instead.
def send_on_queue(queue, obj):
    if not is_pickleable(obj):
        logging.warning("Cannot pickle: {:}. Sending None instead".format(obj))
        queue.put(None)
        return

    if sys.getsizeof(obj) > PIPE_MAX_SEND_BYTES:
        logging.warning(
            "Object size ({:} bytes) exceeds maximum size that can be sent over queues ({:} bytes). Attempting to compress - this may take some time. If this does not work or you want to avoid the compression overhead, you should disable subprocesses via the --no-subprocess flag, or by setting use_subprocess=False in Comparator.run().".format(
                sys.getsizeof(obj), PIPE_MAX_SEND_BYTES
            )
        )
        obj = compress(obj)

    if sys.getsizeof(obj) > PIPE_MAX_SEND_BYTES:
        logging.warning("Compressed object is still too large to send. Sending None instead.")
        queue.put(None)
        return

    logging.info("Sending: {:} on queue".format(obj))
    queue.put(obj)


def receive_on_queue(queue, timeout=None):
    logging.info("Waiting for data to become available on queue")
    obj = queue.get(block=True, timeout=timeout)
    if is_compressed(obj):
        logging.debug("Decompressing output")
        obj = decompress(obj)
    logging.info("Received {:} on queue".format(obj))
    return obj


def timestamped_filepath(dir, name):
    name, ext = os.path.splitext(name)
    return os.path.join(dir, "{:}.{:}{:}".format(name, time.strftime("%Y-%m-%d-%H-%M-%S"), ext))


def write_timestamped(contents, dir=None, name=None, mode="wb"):
    """
    Generates a timestamped file path in the specified directory.

    Args:
        contents (bytes-like object or callable): Either a bytes-like object that can be written to disk, or a callable which will return such an object.
        dir (str): The directory to write into.
        name (str): The name of the file.

    Optional Args:
        mode(str): The mode to use when writing. Defaults to "wb".

    Returns:
        str: The complete file path, or None if nothing was written.
    """
    if dir is not None:
        if not os.path.exists(dir):
            # logging.debug("{:} does not exist, creating now.".format(dir))
            os.makedirs(dir, exist_ok=True)

        path = timestamped_filepath(dir, name)

        if callable(contents):
            contents = contents()

        if os.path.exists(path):
            logging.warning("{:} already exists. Will not overwrite.".format(path))
        else:
            with open(path, mode) as f:
                logging.info("Writing to {:}".format(path))
                f.write(contents)
            return path
    return None


def get_input_metadata_from_profile(profile, network):
    input_metadata = OrderedDict()
    for index in range(network.num_inputs):
        tensor = network.get_input(index)
        if tensor.is_shape_tensor:
            shapes = profile.get_shape_input(tensor.name)
        else:
            shapes = profile.get_shape(tensor.name)
        if tuple(shapes[0]) != tuple(shapes[1]):
            logging.warning("In profile 0, min != max, using opt shapes for calibration")
        # Always use opt shape
        input_metadata[tensor.name] = (trt.nptype(tensor.dtype), shapes[1])
    return input_metadata


class BaseBuffer(object):
    def __init__(self, shape=None, dtype=None):
        self.dtype = default_value(dtype, np.float32)
        self.shape = default_value(shape, tuple())
        self.allocate(self.shape)

    # If the new shape is larger, reallocate, otherwise do nothing.
    def resize(self, new_shape):
        if volume(new_shape) > volume(self.shape):
            self.free()
            self.allocate(new_shape)
        self.shape = new_shape

    def __str__(self):
        return "({:}: shape={:}, dtype={:})".format(type(self).__name__, self.shape, self.dtype)

    def __repr__(self):
        return self.__str__()


class DeviceBuffer(BaseBuffer):
    def allocate(self, shape):
        self.ptr = cuda.mem_alloc(volume(shape) * np.dtype(self.dtype).itemsize)

    def free(self):
        self.ptr.free()

    # Copies a numpy buffer to device
    def copy_htod(self, np_buffer, stream=None):
        if stream:
            # PyCUDA requires the host buffer to be pagelocked for asynchronous memcpys.
            pagelocked = cuda.register_host_memory(np.ascontiguousarray(np_buffer.ravel()))
            cuda.memcpy_htod_async(self.ptr, pagelocked, stream)
        else:
            cuda.memcpy_htod(self.ptr, np.ascontiguousarray(np_buffer.ravel()))


class HostBuffer(BaseBuffer):
    def allocate(self, shape):
        self.ptr = cuda.pagelocked_empty(shape, self.dtype).ravel()

    def free(self):
        del self.ptr

    # Copies a DeviceBuffer to host
    def copy_dtoh(self, device_buffer, stream=None):
        if stream:
            cuda.memcpy_dtoh_async(self.ptr, device_buffer.ptr, stream)
        else:
            cuda.memcpy_dtoh(self.ptr, device_buffer.ptr)

    # Return a view of the buffer which has the correct shape
    def view(self):
        return self.ptr[: volume(self.shape)].reshape(self.shape)


class Buffers(object):
    @staticmethod
    def from_engine(engine):
        buffers = Buffers()
        for binding in engine:
            dtype = trt.nptype(engine.get_binding_dtype(binding))
            buffers.device_buffers[binding] = DeviceBuffer(dtype=dtype)
            if not engine.binding_is_input(binding):
                buffers.host_outputs[binding] = HostBuffer(dtype=dtype)
        return buffers

    def __init__(self):
        self.device_buffers = OrderedDict()
        self.host_outputs = OrderedDict()

    # Resize the specfied buffer to the specified shape
    def resize(self, name, shape):
        found = False
        for buf_dict in [self.device_buffers, self.host_outputs]:
            if name in buf_dict:
                found = True
                buf_dict[name].resize(shape)

        if not found:
            logging.warning("Buffer: {:} was not found, could not resize".format(name))
        else:
            logging.debug("Resizing {:} buffer to {:}".format(name, shape))

    def copy_inputs(self, feed_dict, stream=None):
        for name, buffer in feed_dict.items():
            self.device_buffers[name].copy_htod(buffer, stream)

    # Copies outputs from the device back to host.
    def copy_outputs(self, stream=None):
        for name, buffer in self.host_outputs.items():
            buffer.copy_dtoh(self.device_buffers[name], stream)

    def get_bindings(self):
        return [int(buf.ptr) for buf in self.device_buffers.values()]

    # Gets a dictionary mapping names to numpy buffers.
    def get_outputs(self):
        out_dict = OrderedDict()
        for name, buffer in self.host_outputs.items():
            out_dict[name] = buffer.view()
        return out_dict

    def free(self):
        [buf.free() for buf in self.device_buffers.values()]
        [buf.free() for buf in self.host_outputs.values()]


class BaseRunner(object):
    def __init__(self, name=None):
        """
        The base class for runner objects. All runners should override the functions and attributes specified here.

        Vars:
            name (str): The name of this runner.
        """
        self.name = default_value(name, "Runner")
        self.inference_time = None

    def __enter__(self):
        """
        Activate the runner for inference. This may involve allocating GPU buffers, for example.

        It is extremely important that the GPU is not used by the runner before the __enter__ function is called.

        Vars:
            inputs (OrderedDict[InputKey, Tuple[int]]): A mapping of input tensor names to their shapes, INCLUDING batch dimension, for this runner. This MUST be known at runner initialization for the Comparator to work correctly. InputKey can be any type used to uniquely indentify an input, e.g. a string containing the input name.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Deactivate the runner. All memory allocated by __enter__() should be freed, such that the runner is no longer using the GPU.

        Returns:
            BaseRunner
        """
        pass

    def infer(self, feed_dict, output):
        """
        Runs inference using the provided feed_dict.

        Args:
            feed_dict (OrderedDict[str, np.ndarray]): A mapping of input tensor names to corresponding input numpy arrays.

        Returns:
            OrderedDict[str, np.ndarray]: A mapping of output tensor names to their corresponding numpy arrays.
        """
        raise NotImplementedError("BaseRunner is an abstract class")

    # def last_inference_time(self):
    #     """
    #     Returns the time required for the last call to `infer`
    #
    #     Returns:
    #         float: The time in seconds
    #     """
    #     if self.inference_time is None:
    #         logging.warning("inference_time was not set for this runner. Inference time will be incorrect! To correctly compare runtimes, please set the inference_time property in the infer() function", mode=LogMode.ONCE)
    #         return 1
    #     return self.inference_time

    def get_input_metadata(self, input_examples, output_example):
        """
        Returns information about the inputs of the model. Shapes here may be unknown/dynamic. Must be called after __enter__() and before __exit__()

        Returns:
            OrderedDict[str, Tuple[np.dtype, Tuple[int]]]: Mapping of input names to their data types and shapes.
        """
        raise NotImplementedError("BaseRunner is an abstract class")


# Builds and tracks a single engine for a single network.
class TensorRTRunnerV2(BaseRunner):
    total_runners = 0
    """
    A runner that can perform inference on a single TensorRT engine.
    """

    def __init__(self, model_loader=None, plugins=None, name=None):
        """
        Creates a runner that manages a single TensorRT engine.

        Args:
            model_loader (Callable() -> trt.ICudaEngine): A callable that can supply a TensorRT engine.

        Optional Args:
            max_workspace_size (int): The maximum workspace size in bytes.
            plugins (List[str]): A list of paths to plugin libraries to load before inference.
            name (str): The human-readable name to use for this runner.
        """
        set_trt_logging_level(logging.getEffectiveLevel())

        def load_plugins():
            import ctypes

            for plugin in plugins:
                path = os.path.abspath(plugin)
                logging.info("Loading plugin library: {:}".format(path))
                ctypes.CDLL(path)

        # Load any user-supplied plugin libraries. This must happen before everything else, including engine deserialization.
        if plugins:
            load_plugins()

        # Choose a unique name for this runner.
        super().__init__(default_value(name, "trt-v2-runner-{:}".format(TensorRTRunnerV2.total_runners)))
        TensorRTRunnerV2.total_runners += 1
        logging.debug("Creating {:}".format(self.name))

        self.model_loader = model_loader

        self.engine = self.model_loader()
        if not self.engine:
            logging.critical("Invalid Engine. Please ensure the engine was built correctly.")

        self.buffers = Buffers.from_engine(self.engine)
        self.stream = cuda.Stream()

        self.context = self.engine.create_execution_context()

    def __enter__(self):
        """
        Vars:
            engine (trt.ICudaEngine): The engine tracked by this runner. The TensorRTRunnerV2 OWNS the engine it manages, and therefore is responsible for it's destruction. Do not free the engine outside of the runner, or it will result in a double free.
            context (trt.IExecutionContext): The context used for inference.
            stream (pycuda.driver.Stream): The CUDA stream that this runner will use for inference.
        """
        return self

    @staticmethod
    def override_shape_list(shape):
        return [DEFAULT_SHAPE_VALUE if is_dimension_dynamic(dim) else dim for dim in shape]

    def get_input_metadata(self):
        inputs = OrderedDict()
        active_profile = self.context.active_optimization_profile
        bindings_per_profile = len(self.engine) // self.engine.num_optimization_profiles
        logging.debug(
            "Total # of Profiles: {:}, Bindings Per Profile: {:}, Active Profile: {:}".format(
                self.engine.num_optimization_profiles, bindings_per_profile, active_profile
            )
        )

        start_binding = bindings_per_profile * active_profile
        end_binding = start_binding + bindings_per_profile
        logging.info("Start Binding: {:}, End Binding: {:}".format(start_binding, end_binding))

        for binding in range(start_binding, end_binding):
            if self.engine.binding_is_input(binding):
                inputs[self.engine[binding]] = (
                    trt.nptype(self.engine.get_binding_dtype(binding)),
                    list(self.engine.get_binding_shape(binding)),
                )
        return inputs

    def __exit__(self, exc_type, exc_value, traceback):
        # Destroy the engine, and context.
        with self.engine, self.context:
            pass

        self.buffers.free()
        del self.stream

    def infer(self, feed_dict, output):
        for name in self.engine:
            if name in feed_dict:
                in_out = [feed_dict[name]]
            elif isinstance(output, tuple):
                in_out = [output[i].detach().cpu().numpy() for i in range(len(output))]
            else:
                in_out = [output.detach().cpu().numpy()]

            binding = self.engine[name]

            # Only set shapes if required
            for i in range(len(in_out)):
                shape = in_out[i].shape
                if self.engine.is_shape_binding(binding) and is_shape_dynamic(self.context.get_shape(binding)):
                    logging.debug("Setting shape binding: {:} (index: {:}) to: {:}".format(name, binding, in_out[i]))
                    self.context.set_shape_input(binding, in_out[i])
                elif is_shape_dynamic(self.context.get_binding_shape(binding)):
                    logging.debug("Setting binding: {:} (index: {:}) to shape: {:}".format(name, binding, shape))
                    self.context.set_binding_shape(binding, shape)

        # Check
        if not self.context.all_binding_shapes_specified:
            logging.critical(
                "Some input shapes were not specified.\nNote: Inputs are: {:}".format(self.get_input_metadata())
            )
        if not self.context.all_shape_inputs_specified:
            logging.critical(
                "Some shape inputs were not specified.\nNote: Inputs are: {:}".format(self.get_input_metadata())
            )

        bindings_per_profile = self.engine.num_bindings // self.engine.num_optimization_profiles
        start_binding = self.context.active_optimization_profile * bindings_per_profile
        end_binding = start_binding + bindings_per_profile

        # Resize buffers so they are the appropriate size.
        for binding in range(start_binding, end_binding):
            shape = tuple(self.context.get_binding_shape(binding))
            self.buffers.resize(self.engine[binding], shape)

        bindings = self.buffers.get_bindings()

        start = time.perf_counter()
        self.buffers.copy_inputs(feed_dict, self.stream)
        self.context.execute_async_v2(bindings=bindings, stream_handle=self.stream.handle)
        self.buffers.copy_outputs(self.stream)
        self.stream.synchronize()
        end = time.perf_counter()

        self.inference_time = end - start
        return self.buffers.get_outputs()

    # Utility functions related to TensorRT, but not tied to any specific instance.
    @staticmethod
    def create_network(explicit_batch=True, explicit_precision=False):
        with trt.Builder(TRT_LOGGER) as builder:
            network_flags = 0
            if explicit_batch:
                network_flags = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
            if explicit_precision:
                network_flags = network_flags | (1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_PRECISION))
            network = builder.create_network(flags=network_flags)
            if network is None:
                logging.critical("Invalid network")
            return network

    @staticmethod
    def get_network_inputs(network):
        return {network.get_input(i).name: network.get_input(i).shape for i in range(network.num_inputs)}

    @staticmethod
    def log_network(network):
        LAYER_TYPE_CLASS_MAPPING = {
            trt.LayerType.CONVOLUTION: trt.IConvolutionLayer,
            trt.LayerType.FULLY_CONNECTED: trt.IFullyConnectedLayer,
            trt.LayerType.ACTIVATION: trt.IActivationLayer,
            trt.LayerType.POOLING: trt.IPoolingLayer,
            trt.LayerType.LRN: trt.ILRNLayer,
            trt.LayerType.SCALE: trt.IScaleLayer,
            trt.LayerType.SOFTMAX: trt.ISoftMaxLayer,
            trt.LayerType.DECONVOLUTION: trt.IDeconvolutionLayer,
            trt.LayerType.CONCATENATION: trt.IConcatenationLayer,
            trt.LayerType.ELEMENTWISE: trt.IElementWiseLayer,
            trt.LayerType.PLUGIN: trt.IPluginLayer,
            trt.LayerType.RNN: trt.IRNNLayer,
            trt.LayerType.UNARY: trt.IUnaryLayer,
            trt.LayerType.PADDING: trt.IPaddingLayer,
            trt.LayerType.SHUFFLE: trt.IShuffleLayer,
            trt.LayerType.REDUCE: trt.IReduceLayer,
            trt.LayerType.TOPK: trt.ITopKLayer,
            trt.LayerType.GATHER: trt.IGatherLayer,
            trt.LayerType.MATRIX_MULTIPLY: trt.IMatrixMultiplyLayer,
            trt.LayerType.RAGGED_SOFTMAX: trt.IRaggedSoftMaxLayer,
            trt.LayerType.CONSTANT: trt.IConstantLayer,
            trt.LayerType.RNN_V2: trt.IRNNv2Layer,
            trt.LayerType.IDENTITY: trt.IIdentityLayer,
            trt.LayerType.PLUGIN_V2: trt.IPluginV2Layer,
            trt.LayerType.SLICE: trt.ISliceLayer,
            trt.LayerType.SHAPE: trt.IShapeLayer,
            trt.LayerType.PARAMETRIC_RELU: trt.IParametricReLULayer,
            trt.LayerType.RESIZE: trt.IResizeLayer,
        }

        def is_special_attribute(attr):
            return attr.startswith("__") and attr.endswith("__")

        def is_valid_attribute(attr, layer):
            if (
                type(layer) == trt.IPoolingLayer
                or type(layer) == trt.IConvolutionLayer
                or type(layer) == trt.IDeconvolutionLayer
            ):
                if len(layer.get_input(0).shape) > 4:
                    # 3D pooling uses padding_nd
                    return attr not in ["padding", "stride", "window_size"]
            if type(layer) == trt.IResizeLayer:
                if layer.num_inputs > 1:
                    return attr not in ["scales"]
            if type(layer) == trt.ISliceLayer:
                if layer.num_inputs > 1:
                    return attr not in ["shape", "start", "stride"]
            return True

        logging.debug("Network Inputs: {:}".format(TensorRTRunnerV2.get_network_inputs(network)))
        for layer in network:
            if layer.type in LAYER_TYPE_CLASS_MAPPING:
                layer.__class__ = LAYER_TYPE_CLASS_MAPPING[layer.type]
            input_info = [
                "{:}: {:} ({:})".format(layer.get_input(i).name, layer.get_input(i).shape, layer.get_input(i).dtype)
                for i in range(layer.num_inputs)
                if layer.get_input(i)
            ]
            output_info = [
                "{:}: {:} ({:})".format(layer.get_output(i).name, layer.get_output(i).shape, layer.get_output(i).dtype)
                for i in range(layer.num_outputs)
                if layer.get_output(i)
            ]
            logging.info("{:} [Op: {:}]".format(layer.name, layer.type))
            logging.info("\t{:} -> {:}".format(input_info, output_info))
            attrs = dir(layer)
            for attr in attrs:
                # Exclude special attributes, as well as any attributes of the base layer class (those can be displayed above).
                if (
                    not is_special_attribute(attr)
                    and not hasattr(trt.ILayer, attr)
                    and is_valid_attribute(attr, layer)
                ):
                    logging.info("\t{:}.{:} = {:}".format(layer.name, attr, getattr(layer, attr)))

        network_outputs = {network.get_output(i).name: network.get_output(i).shape for i in range(network.num_outputs)}
        logging.debug("Network Outputs: {:}".format(network_outputs))

    @staticmethod
    def mark_layerwise(network):
        # Layers within loops cannot be marked as network outputs.
        # TODO: FIXME: This assumes that the network is topologically sorted.
        LOOP_START_LAYERS = [trt.LayerType.TRIP_LIMIT, trt.LayerType.ITERATOR]
        LOOP_END_LAYERS = [trt.LayerType.LOOP_OUTPUT]
        num_layers_marked = 0
        in_loop = False
        for layer in network:
            if layer.type in LOOP_START_LAYERS:
                in_loop = True
            elif layer.type in LOOP_END_LAYERS:
                in_loop = False
            for index in range(layer.num_outputs):
                out = layer.get_output(index)
                if not out.is_network_output and not in_loop:
                    logging.debug("Marking {:} as an output".format(out.name))
                    network.mark_output(out)
                    num_layers_marked += 1
        logging.debug("Running in layerwise mode. Marking {:} layers as outputs".format(num_layers_marked))

    @staticmethod
    def build_profile(builder, network, profile_shapes, default_shape_value=DEFAULT_SHAPE_VALUE):
        def override_shape(shape):
            return tuple([DEFAULT_SHAPE_VALUE if is_dimension_dynamic(dim) else dim for dim in shape])

        def get_profile_shape(name):
            if name not in profile_shapes:
                return None
            shapes = profile_shapes[name]
            if not isinstance(shapes, list) or len(shapes) != 3:
                logging.critical(
                    "Profile values must be a list containing exactly 3 shapes (tuples or Dims), but received shapes: {:} for input: {:}.\nNote: profile was: {:}.\nNote: Network inputs were: {:}".format(
                        shapes, name, profile_shapes, TensorRTRunnerV2.get_network_inputs(network)
                    )
                )
            return shapes

        profile = builder.create_optimization_profile()
        for idx in range(network.num_inputs):
            inp = network.get_input(idx)

            if inp.is_shape_tensor:
                shapes = get_profile_shape(inp.name)
                if not shapes:
                    rank = inp.shape[0]
                    shapes = [(DEFAULT_SHAPE_VALUE,) * rank] * 3
                    logging.warning(
                        "Setting shape input to {:}. If this is incorrect, for shape input: {:}, please provide tuples for min, opt, and max shapes containing {:} elements".format(
                            shapes[0], inp.name, rank
                        ),
                        mode=logging_mode.ONCE,
                    )
                min, opt, max = shapes
                profile.set_shape(inp.name, min, opt, max)
                inp.shape = opt
                logging.info(
                    "Setting shape input: {:} values to min: {:}, opt: {:}, max: {:}".format(inp.name, min, opt, max)
                )
            else:
                shapes = get_profile_shape(inp.name)
                if not shapes:
                    shapes = [override_shape(inp.shape)] * 3
                    logging.warning(
                        "Overriding input shape {:} to {:}. If this is incorrect, for input tensor: {:}, please provide tuples for min, opt, and max shapes containing values: {:} with dynamic dimensions replaced,".format(
                            inp.shape, shapes[0], inp.name, inp.shape
                        ),
                        mode=logging_mode.ONCE,
                    )
                min, opt, max = shapes
                profile.set_shape(inp.name, min, opt, max)
                inp.shape = opt
                logging.info(
                    "Setting input: {:} shape to min: {:}, opt: {:}, max: {:}".format(inp.name, min, opt, max)
                )

        if not profile:
            logging.critical(
                "Profile is not valid, please provide profile data. Note: profile was: {:}".format(profile_shapes)
            )
        return profile



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from numba import jit


def phase_vocoder(D: np.ndarray, rate: float, phi_advance: np.ndarray, scale_buffer: np.ndarray):
    """
    Optimized implementation of phase vocoder from Librosa.

    Reference implementation:
        - https://librosa.github.io/librosa/generated/librosa.core.phase_vocoder.html

    Args:
        D: Complex spectograms of shape [d, t, complex=2].
        rate: Speed rate, must be float greater than 0.
        phi_advance: Precomputed phase advance buffer array of length [n_fft + 1]
        scale_buffer: Precomputed numpy buffer array of length [n_fft + 1]

    Returns:
        Complex64 ndarray of shape [d, t / rate, complex=2]
    """
    time_steps = np.arange(0, D.shape[1], rate, dtype=np.float)

    # Create an empty output array
    d_stretch = np.zeros((D.shape[0], len(time_steps)), D.dtype, order='F')

    # Phase accumulator; initialize to the first sample
    phase_acc = np.angle(D[:, 0])

    # Pad 0 columns to simplify boundary logic
    D = np.pad(D, [(0, 0), (0, 2)], mode='constant')

    d_stretch = _phase_vocoder_kernel(D, time_steps, phi_advance, d_stretch, phase_acc, scale_buffer)

    return d_stretch


@jit(nopython=True, nogil=True)
def _phase_vocoder_kernel(D, time_steps, phi_advance, d_stretch, phase_acc, scale_buffer):
    """
    Numba optimized kernel to compute the phase vocoder step.

    Args:
        D: Complex spectograms of shape [d, t, complex=2].
        rate: Speed rate, must be float greater than 0.
        time_steps: Numpy ndarray of linearly spaced time steps, shape = [t]
        phi_advance: Precomputed phase advance buffer array of length [n_fft + 1]
        d_stretch: Output complex matrix of shape [d, t / rate, complex=2]
        phase_acc: Phase accumulator initialized to first sample of shape [d, complex=2]
        scale_buffer: Precomputed numpy buffer array of length [n_fft + 1]

    Returns:
        Complex64 ndarray of shape [d, t / rate, complex=2]
    """
    two_pi = 2.0 * np.pi

    for (t, step) in enumerate(time_steps):
        columns = D[:, int(step) : int(step + 2)]
        columns_0 = columns[:, 0]
        columns_1 = columns[:, 1]

        # Weighting for linear magnitude interpolation
        alpha = np.mod(step, 1.0)
        mag = (1.0 - alpha) * np.abs(columns_0) + alpha * np.abs(columns_1)

        # Store to output array
        d_stretch[:, t] = mag * np.exp(1.0j * phase_acc)

        # Compute phase advance
        dphase = np.angle(columns_1) - np.angle(columns_0) - phi_advance

        # Wrap to -pi:pi range
        scale = dphase / two_pi
        np.round(scale, 0, scale_buffer)

        dphase = dphase - two_pi * scale_buffer

        # Accumulate phase
        phase_acc += phi_advance + dphase

    return d_stretch



# Copyright (c) 2019 NVIDIA Corporation
from functools import partial

import torch
from torch import nn
from torch.nn.functional import pad

from nemo import logging
from nemo.backends.pytorch.nm import LossNM, NonTrainableNM, TrainableNM
from nemo.collections.tts.parts.layers import get_mask_from_lengths
from nemo.collections.tts.parts.tacotron2 import Decoder, Encoder, Postnet
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs

__all__ = [
    "MakeGate",
    "Tacotron2Loss",
    "Tacotron2Postnet",
    "Tacotron2Decoder",
    "Tacotron2DecoderInfer",
    "Tacotron2Encoder",
    "TextEmbedding",
]


class TextEmbedding(TrainableNM):
    """
    TextEmbedding embeds the encoded character labels to an embedding space

    Args:
        n_symbols (int): The number of character labels. The input char_phone's
            second axis dim size should be n_symbols.
        symbols_embedding_dim (int): The size of the embedding dimension.
            Defaults to 512.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        # return {"char_phone": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)})}
        return {"char_phone": NeuralType(('B', 'T'), LabelsType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "char_phone_embeddings": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(EmbeddedTextTag), 2: AxisType(TimeTag),}
            # )
            "char_phone_embeddings": NeuralType(('B', 'D', 'T'), EmbeddedTextType())
        }

    def __init__(self, n_symbols, symbols_embedding_dim: int = 512):
        super().__init__()
        self.embedding = nn.Embedding(n_symbols, symbols_embedding_dim)
        self.to(self._device)

    def forward(self, char_phone):
        return self.embedding(char_phone).transpose(1, 2)


class Tacotron2Encoder(TrainableNM):
    """
    Tacotron2Encoder is the encoder part of Tacotron 2. It takes embedded text
    as input and creates an encoded representation of the text that can be used
    with downstream attention and decoders.

    Args:
        encoder_n_convolutions (int): The number of convolution layers inside
            the encoder. Defaults to 5.
        encoder_embedding_dim (int): The size of the embedded text. It will
            also be the output size of the encoded text. Defaults to 512.
        encoder_kernel_size (int): The kernel size of the convolution layers.
            Defaults to 3
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "char_phone_embeddings": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(EmbeddedTextTag), 2: AxisType(TimeTag),}
            # ),
            # "embedding_length": NeuralType({0: AxisType(BatchTag)}),
            "char_phone_embeddings": NeuralType(('B', 'D', 'T'), EmbeddedTextType()),
            "embedding_length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "char_phone_encoded": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(EncodedRepresentationTag),}
            # )
            "char_phone_encoded": NeuralType(('B', 'T', 'D'), EncodedRepresentation())
        }

    def __init__(
        self, encoder_n_convolutions: int = 5, encoder_embedding_dim: int = 512, encoder_kernel_size: int = 3
    ):
        super().__init__()
        self.encoder = Encoder(
            encoder_n_convolutions=encoder_n_convolutions,
            encoder_embedding_dim=encoder_embedding_dim,
            encoder_kernel_size=encoder_kernel_size,
        )
        self.to(self._device)

    def forward(self, char_phone_embeddings, embedding_length):
        char_phone_encoded = self.encoder(char_phone_embeddings, embedding_length)
        return char_phone_encoded


class Tacotron2Decoder(TrainableNM):
    """
    Tacotron2Decoder implements the attention, decoder, and prenet parts of
    Tacotron 2. It takes the encoded text and produces mel spectrograms. The
    decoder contains two rnns, one is called the decoder rnn and the other is
    called the attention rnn.

    Args:
        n_mel_channels (int): The size or dimensionality of the mel spectrogram
        n_frames_per_step (int): The number of frames we predict at each
            decoder time step. Defaults to 1
        encoder_embedding_dim (int): The size of the encoded text.
            Defaults to 512.
        gate_threshold (float): A number in [0, 1). When teacher forcing is
            not used, the model predict a stopping value at each model time
            step. The model will stop if the value is greater than
            gate_threshold. Defaults to 0.5.
        prenet_dim (int): The hidden dimension of the prenet. Defaults to 256.
        max_decoder_steps (int): When not teacher forcing, the maximum number
            of frames to predict. Defaults to 1000.
        decoder_rnn_dim (int): The hidden dimension of the decoder rnn.
            Defaults to 1024.
        p_decoder_dropout (float): Dropout probability for the decoder rnn.
            Defaults to 0.1.
        p_attention_dropout (float): Dropout probability for the attention rnn.
            Defaults to 0.1.
        attention_rnn_dim (int): The hidden dimension of the attention rnn.
            Defaults to 1024.
        attention_dim (int): The hidden dimension of the attention mechanism.
            Defaults to 128.
        attention_location_n_filters (int): The number of convolution filters
            for the location part of the attention mechanism.
            Defaults to 32.
        attention_location_kernel_size (int): The kernel size of the
            convolution for the location part of the attention mechanism.
            Defaults to 31.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "char_phone_encoded": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(EncodedRepresentationTag),}
            # ),
            # "encoded_length": NeuralType({0: AxisType(BatchTag)}),
            # "mel_target": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            "char_phone_encoded": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
            "encoded_length": NeuralType(tuple('B'), LengthsType()),
            "mel_target": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "mel_output": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            # "gate_output": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "alignments": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(TimeTag),}),
            "mel_output": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "gate_output": NeuralType(('B', 'T'), ChannelType()),
            "alignments": NeuralType(('B', 'T', 'T'), ChannelType()),
        }

    def __init__(
        self,
        n_mel_channels: int,
        n_frames_per_step: int = 1,
        encoder_embedding_dim: int = 512,
        gate_threshold: float = 0.5,
        prenet_dim: int = 256,
        max_decoder_steps: int = 1000,
        decoder_rnn_dim: int = 1024,
        p_decoder_dropout: float = 0.1,
        p_attention_dropout: float = 0.1,
        attention_rnn_dim: int = 1024,
        attention_dim: int = 128,
        attention_location_n_filters: int = 32,
        attention_location_kernel_size: int = 31,
        prenet_p_dropout: float = 0.5,
        force: bool = False,
    ):
        super().__init__()
        self.decoder = Decoder(
            n_mel_channels=n_mel_channels,
            n_frames_per_step=n_frames_per_step,
            encoder_embedding_dim=encoder_embedding_dim,
            gate_threshold=gate_threshold,
            prenet_dim=prenet_dim,
            max_decoder_steps=max_decoder_steps,
            decoder_rnn_dim=decoder_rnn_dim,
            p_decoder_dropout=p_decoder_dropout,
            p_attention_dropout=p_attention_dropout,
            attention_rnn_dim=attention_rnn_dim,
            attention_dim=attention_dim,
            attention_location_n_filters=attention_location_n_filters,
            attention_location_kernel_size=attention_location_kernel_size,
            prenet_p_dropout=prenet_p_dropout,
            early_stopping=True,
        )
        self.force = force
        self.to(self._device)

    def forward(self, char_phone_encoded, encoded_length, mel_target):
        if self.training or self.force:
            mel_output, gate_output, alignments = self.decoder(
                char_phone_encoded, mel_target, memory_lengths=encoded_length
            )
        else:
            mel_output, gate_output, alignments, _ = self.decoder.infer(
                char_phone_encoded, memory_lengths=encoded_length
            )
        return mel_output, gate_output, alignments


class Tacotron2DecoderInfer(Tacotron2Decoder):
    """
    Tacotron2DecoderInfer is an inference Neural Module used in place
    of the Tacotron2Decoder NM.

    Args:
        n_mel_channels (int): The size or dimensionality of the mel spectrogram
        n_frames_per_step (int): The number of frames we predict at each
            decoder time step. Defaults to 1
        encoder_embedding_dim (int): The size of the encoded text.
            Defaults to 512.
        gate_threshold (float): A number in [0, 1). When teacher forcing is
            not used, the model predict a stopping value at each model time
            step. The model will stop if the value is greater than
            gate_threshold. Defaults to 0.5.
        prenet_dim (int): The hidden dimension of the prenet. Defaults to 256.
        max_decoder_steps (int): When not teacher forcing, the maximum number
            of frames to predict. Defaults to 1000.
        decoder_rnn_dim (int): The hidden dimension of the decoder rnn.
            Defaults to 1024.
        p_decoder_dropout (float): Dropout probability for the decoder rnn.
            Defaults to 0.1.
        p_attention_dropout (float): Dropout probability for the attention rnn.
            Defaults to 0.1.
        attention_rnn_dim (int): The hidden dimension of the attention rnn.
            Defaults to 1024.
        attention_dim (int): The hidden dimension of the attention mechanism.
            Defaults to 128.
        attention_location_n_filters (int): The number of convolution filters
            for the location part of the attention mechanism.
            Defaults to 32.
        attention_location_kernel_size (int): The kernel size of the
            convolution for the location part of the attention mechanism.
            Defaults to 31.
    """

    def __init__(
        self,
        n_mel_channels: int,
        n_frames_per_step: int = 1,
        encoder_embedding_dim: int = 512,
        gate_threshold: float = 0.5,
        prenet_dim: int = 256,
        max_decoder_steps: int = 1000,
        decoder_rnn_dim: int = 1024,
        p_decoder_dropout: float = 0.1,
        p_attention_dropout: float = 0.1,
        attention_rnn_dim: int = 1024,
        attention_dim: int = 128,
        attention_location_n_filters: int = 32,
        attention_location_kernel_size: int = 31,
        prenet_p_dropout: float = 0.5,
        force: bool = False,
    ):
        super().__init__(
            n_mel_channels=n_mel_channels,
            n_frames_per_step=n_frames_per_step,
            encoder_embedding_dim=encoder_embedding_dim,
            gate_threshold=gate_threshold,
            prenet_dim=prenet_dim,
            max_decoder_steps=max_decoder_steps,
            decoder_rnn_dim=decoder_rnn_dim,
            p_decoder_dropout=p_decoder_dropout,
            p_attention_dropout=p_attention_dropout,
            attention_rnn_dim=attention_rnn_dim,
            attention_dim=attention_dim,
            attention_location_n_filters=attention_location_n_filters,
            attention_location_kernel_size=attention_location_kernel_size,
            prenet_p_dropout=prenet_p_dropout,
            force=force,
        )

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "char_phone_encoded": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(EncodedRepresentationTag),}
            # ),
            # "encoded_length": NeuralType({0: AxisType(BatchTag)}),
            "char_phone_encoded": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
            "encoded_length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "mel_output": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            # "gate_output": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "alignments": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(TimeTag),}),
            # "mel_len": NeuralType({0: AxisType(BatchTag)}),
            "mel_output": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "gate_output": NeuralType(('B', 'T'), ChannelType()),
            "alignments": NeuralType(('B', 'T', 'T'), ChannelType()),
            "mel_len": NeuralType(tuple('B'), LengthsType()),
        }

    def __str__(self):
        return "Tacotron2Decoder"

    def forward(self, char_phone_encoded, encoded_length):
        if self.training:
            raise ValueError("You are using the Tacotron 2 Infer Neural Module in training mode.")
        with torch.no_grad():
            mel_output, gate_output, alignments, mel_len = self.decoder.infer(
                char_phone_encoded, memory_lengths=encoded_length
            )
        return mel_output, gate_output, alignments, mel_len


class Tacotron2Postnet(TrainableNM):
    """
    Tacotron2Postnet implements the postnet part of Tacotron 2. It takes a mel
    spectrogram as generated by the decoder and corrects errors within the
    generated mel spectrogram.

    Args:
        n_mel_channels (int): The size or dimensionality of the mel spectrogram
        postnet_embedding_dim (int): Hidden size of convolutions.
            Defaults to 512.
        postnet_kernel_size (int): Kernel size of convolutions.
            Defaults to 5.
        postnet_n_convolutions (int): Number of convolution layers.
            Defaults to 5.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "mel_input": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # )
            "mel_input": NeuralType(('B', 'D', 'T'), MelSpectrogramType())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "mel_output": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            "mel_output": NeuralType(('B', 'D', 'T'), MelSpectrogramType())
        }

    def __init__(
        self,
        n_mel_channels: int,
        postnet_embedding_dim: int = 512,
        postnet_kernel_size: int = 5,
        postnet_n_convolutions: int = 5,
        p_dropout: float = 0.5,
    ):
        super().__init__()
        self.postnet = Postnet(
            n_mel_channels=n_mel_channels,
            postnet_embedding_dim=postnet_embedding_dim,
            postnet_kernel_size=postnet_kernel_size,
            postnet_n_convolutions=postnet_n_convolutions,
            p_dropout=p_dropout,
        )
        self.to(self._device)

    def forward(self, mel_input):
        return self.postnet(mel_input) + mel_input


class Tacotron2Loss(LossNM):
    """
    Tacoton2Loss implements the loss function of Tacotron 2. The loss function
    is the mean squared error between the reference mel spectrogram and the
    mel spectrogram predicted by the decoder + the mean squared error between
    the reference mel spectrogram and the mel spectrogram predicted by the
    post net + the cross entropy error between the stop values and the
    reference mel length.

    Args:
        pad_value (float): In the evaluation case, when we don't use teacher
            forcing, if the generated mel is shorter than the reference mel,
            we pad the generated mel with this value. Default is ~log(1e-5).
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "mel_out": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            # "mel_out_postnet": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            # "gate_out": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "mel_target": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            # "gate_target": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "target_len": NeuralType({0: AxisType(BatchTag)}),
            # "seq_len": NeuralType({0: AxisType(BatchTag)}),
            "mel_out": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "mel_out_postnet": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "gate_out": NeuralType(('B', 'T'), ChannelType()),
            "mel_target": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "gate_target": NeuralType(('B', 'T'), ChannelType()),
            "target_len": NeuralType(tuple('B'), LengthsType()),
            "seq_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, pad_value: float = -11.52):
        super().__init__()
        self.pad_value = pad_value

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))

    def _loss(
        self, mel_out, mel_out_postnet, gate_out, mel_target, gate_target, target_len, seq_len,
    ):
        mel_target.requires_grad = False
        gate_target.requires_grad = False
        gate_target = gate_target.view(-1, 1)

        max_len = mel_target.shape[2]

        if max_len < mel_out.shape[2]:
            # Predicted len is larger than reference
            # Need to slice
            mel_out = mel_out.narrow(2, 0, max_len)
            mel_out_postnet = mel_out_postnet.narrow(2, 0, max_len)
            gate_out = gate_out.narrow(1, 0, max_len).contiguous()
        elif max_len > mel_out.shape[2]:
            # Need to do padding
            pad_amount = max_len - mel_out.shape[2]
            mel_out = pad(mel_out, (0, pad_amount), value=self.pad_value)
            mel_out_postnet = pad(mel_out_postnet, (0, pad_amount), value=self.pad_value)
            gate_out = pad(gate_out, (0, pad_amount), value=1e3)
            max_len = mel_out.shape[2]

        mask = ~get_mask_from_lengths(target_len, max_len=max_len)
        mask = mask.expand(mel_target.shape[1], mask.size(0), mask.size(1))
        mask = mask.permute(1, 0, 2)
        mel_out.data.masked_fill_(mask, self.pad_value)
        mel_out_postnet.data.masked_fill_(mask, self.pad_value)
        gate_out.data.masked_fill_(mask[:, 0, :], 1e3)

        gate_out = gate_out.view(-1, 1)
        mel_loss = nn.MSELoss()(mel_out, mel_target) + nn.MSELoss()(mel_out_postnet, mel_target)
        gate_loss = nn.BCEWithLogitsLoss()(gate_out, gate_target)
        return mel_loss + gate_loss


class MakeGate(NonTrainableNM):
    """MakeGate is a helper Neural Module that makes the target stop value.
    """

    def __init__(self):
        super().__init__()

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "target_len": NeuralType({0: AxisType(BatchTag)}),
            # "mel_target": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(TimeTag),}
            # ),
            "target_len": NeuralType(tuple('B'), LengthsType()),
            "mel_target": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"gate_target": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)})}
        return {"gate_target": NeuralType(('B', 'T'), ChannelType())}

    def forward(self, target_len, mel_target):
        max_len = mel_target.shape[2]
        gate_padded = torch.FloatTensor(target_len.shape[0], max_len)
        gate_padded.zero_()
        for i, length in enumerate(target_len):
            gate_padded[i, length.data - 1 :] = 1
        return gate_padded.to(device=self._device)



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


from typing import List, Optional

import torch

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.core.neural_types import AxisKind, AxisType, LogprobsType, NeuralType, VoidType
from nemo.utils import logging
from nemo.utils.configuration_parsing import get_value_from_dictionary
from nemo.utils.decorators import add_port_docs

__all__ = ['NonLinearity']


class NonLinearity(NonTrainableNM):
    """
    Class responsible for applying additional non-linearity along the last axis of the input tensor.

    """

    def __init__(self, type: str = "logsoftmax", sizes: List[int] = [-1], name: Optional[str] = None):
        """
        Constructor initializing the  non-linearity.

        Args:
            type: Type of non-linearity (currently only logsoftmax is supported)
            sizes: Sizes of dimensions of the input/output tensors (DEFAULT: [-1] - variable size batch)
            name: Name of the module (DEFAULT: None)
        """
        # Call constructor of parent classes.
        NonTrainableNM.__init__(self, name=name)

        # Store params.
        self._type = type
        self._sizes = sizes

        # Get type - only one option accepted (for now).
        self._non_linearity_type = get_value_from_dictionary(type, ["logsoftmax"])

        # Apply the non-linearity along the last dimension.
        dim = len(sizes) - 1
        self._non_linearity = torch.nn.LogSoftmax(dim=dim)

    @property
    @add_port_docs()
    def input_ports(self):
        """
        Returns definitions of module input ports.
        Batch of inputs, each represented as index [BATCH_SIZE x ... x INPUT_SIZE]
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        for size in self._sizes[1:]:
            axes.append(AxisType(kind=AxisKind.Any, size=size))
        # Return neural type.
        return {"inputs": NeuralType(axes, VoidType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """
        Returns definitions of module output ports.
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        for size in self._sizes[1:]:
            axes.append(AxisType(kind=AxisKind.Any, size=size))
        # Return neural type.
        # TODO: if self._type != "logsoftmax"
        return {"outputs": NeuralType(axes, LogprobsType())}

    def forward(self, inputs):
        """
        Encodes "inputs" in the format of a single tensor.
        Stores reshaped tensor in "outputs" field of in data_streams.

        Args:
            inputs: a tensor [BATCH_SIZE x ...]

        Returns:
            Outputs a tensor [BATCH_SIZE x ...] 
        """
        # print("{}: input shape: {}, device: {}\n".format(self.name, inputs.shape, inputs.device))

        # Reshape.
        # TODO: if self._type != "logsoftmax"
        return self._non_linearity(inputs)



# Copyright (C) NVIDIA CORPORATION. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.****

import numpy as np

# Supported Numpy DTypes: `np.sctypes`
ACCEPTED_INT_NUMBER_FORMATS = (
    int,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
)

ACCEPTED_FLOAT_NUMBER_FORMATS = (
    float,
    np.float,
    np.float16,
    np.float32,
    np.float64,
    np.float128,
)

ACCEPTED_STR_NUMBER_FORMATS = (
    str,
    np.str,
)

ACCEPTED_NUMBER_FORMATS = ACCEPTED_INT_NUMBER_FORMATS + ACCEPTED_FLOAT_NUMBER_FORMATS + ACCEPTED_STR_NUMBER_FORMATS

# NEMO_ENV_VARNAME_DEBUG_VERBOSITY = "NEMO_DEBUG_VERBOSITY"
NEMO_ENV_VARNAME_ENABLE_COLORING = "NEMO_ENABLE_COLORING"
NEMO_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR = "NEMO_REDIRECT_LOGS_TO_STDERR"
NEMO_ENV_VARNAME_TESTING = "NEMO_TESTING"
# NEMO_ENV_VARNAME_SAVE_LOGS_TO_DIR        = "NEMO_SAVE_LOGS_TO_DIR"
